﻿'=================================================================================================================
' * Compiled with Visual Studio 2010 Visual Basic
' * Copyright 2013 Microsoft Inc. All Rights Reserved.
' *
' * Please refer to Microsoft's licensing agreements for object code
' * produced with Visual Studio 2010.
' *
' * Unless required by applicable law or agreed to in writing, software
' * distributed under the License is distributed on an "AS IS" BASIS,
' * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
' * See the License for the specific language governing permissions and
' * limitations under the License.
' *
'
'
'               What: CamBam Printing Plugin  02/12/2014
'                For: CamBam by HexRay, LTD., Versions 9.8N through 9.8P (perhaps beyond)
'               From: Automation Machine Design, LLC
'                 By: Lloyd E. Sponenburgh
'
' 2/02/2014     V1.0.0.0  - crude, but works for proof of concept
' 2/08/2014     V2.0.0.0  - functioning with print preview and scaling/rotation
' 2/12/2014     V3.0.0.0  - working with above, and user selection menu for use of features, plus color changes
' 2/14/2014     V3.0.2.2  - fixed bug created by Microsoft update of .net3.5 code on 02/13/2014
'                           resolved on 2/14/2014 by putting an entry counter in the form1.print event handler
' 2/15/2014     V3.1.1.3  - adding print selection dialog, and tracking down a graphics resource problem with a few users
'                           David (dh42) found this one.  Cb's RefreshView does not work reliably.  UpdateViewport does
' 2/16/2014     V3.2.0.0  - added monochrome printing and 'sticky' config settings via a config file in CB{$common}
'
'
'               Please use freely, but credit me and Automation Machine Design, LLC
'               with high praise and big money for any of the unique routines and/or methods, and for any
'               copies of the plugin.  Please send donations in minimum increments of $1000 per monthly installment.
'
'   known problems:  One late bound resolution of an object at line 406 of this file.  I don't know enough about structured
'   languages yet (being an old procedural guy) to cure it.  It works, but might cause runtime errors in future.
'
'
'
'=================================================================================================================
Imports CamBam
Imports CamBam.CAD
Imports CamBam.Geom
Imports CamBam.UI
Imports CamBam.Util
Imports CamBam.ThisApplication
Imports System
Imports System.Text
Imports System.Windows.Forms
Imports System.Drawing
Imports System.Drawing.Printing
Imports System.Drawing.Font



Public Class Form1
    '
    '
    Dim Logging As Boolean = False      ' set this false to turn off diagnostic logging

    Dim screencap_path As String        ' path to the raw (full color) screen capture file
    Dim selectcap_path As String        ' path to the modified, rotated, scaled, and color-changed 'select' file.

    Dim ui As CamBamUI = CamBam.UI.CamBamUI.MainUI()
    Dim view As CamBam.UI.ICADView = ui.ActiveView



    Dim ourlayers As LayerCollection = view.CADFile.Layers
    Dim layercount As Integer
    Dim colormatrix(layercount) As System.Drawing.Color     ' storage for all the layer colors
    Dim Penmatrix(layercount) As Integer                    ' storage for all the layer penwidths
    Dim Our_penwidth As Integer = 2
    Dim panecolor As System.Drawing.Color       ' value for background color
    Dim logstring As String                     ' string for logging


    Private _bRubberBandingOn As Boolean = False '-- State to control if we are drawing the rubber banding object
    Private _pClickStart As New Point '-- The place where the mouse button went 'down'.
    Private _pClickStop As New Point '-- The place where the mouse button went 'up'.
    Private _pNow As New Point '-- Holds the current mouse location to make the shape appear to follow the mouse cursor.


    Dim _rRectangle As New Rectangle    ' The rectangle used by .NET to get the draw area

    Dim newdoc As Object = view.CADFile ' This view
    Dim p As String = ""
    Dim rawpath As String
    Dim PrintFlag As Integer
    Dim max_window_width As Integer
    Dim max_window_height As Integer
    Dim waittime As Integer = 0
    Dim firstprint As Integer = 0

    ' =-= start code

    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ' construct the form1 for display

        getCBpath()                 ' find and build the path to the capture file
        ' 
        Me.Location = New Point(Screen.PrimaryScreen.Bounds.Left, Screen.PrimaryScreen.Bounds.Top) ' draws the form as big as the screen
        Me.Height = Screen.PrimaryScreen.Bounds.Height + 20  ' and at the same origins as the screen
        Me.Width = Screen.PrimaryScreen.Bounds.Width + 20 ' then below, makes it almost transparent, but still able to be drawn upon

        Me.FormBorderStyle = BorderStyle.None
        Me.ControlBox = False
        Me.Opacity = 0.3

form1_Load_oops:
    End Sub


    Private Sub GetScreenshot_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles GetScreenshot.Click
        ' enter here if the Enter key has been pressed to capture the selected image

        Dim graph As Graphics = Nothing
        Dim result As DialogResult

        Me.Hide()         'Get rid of the overlying form, so it doesn't haze the screen

        ' establishes the boundaries of the whole screen, plus just a little,
        ' to make sure the cursor position and right edge are kept
        Dim bmp As New Bitmap(Screen.PrimaryScreen.Bounds.Width + 5, Screen.PrimaryScreen.Bounds.Height + 5)

        If Logging Then AddLogMessage(4, "CBPrint- preparing to switch BG to white")
        Swapto_white()          ' correct any non-white background for printing, if selected by the user in form2
        If Logging Then AddLogMessage(4, "CBPrint- switch to white finished")
    
        ' a try clause, because the creation of the files might fail
        Try
            graph = Graphics.FromImage(bmp)            'creates a graphic space in user memory
            graph.CopyFromScreen(0, 0, 0, 0, bmp.Size)  ' captures the full screen from which we'll crop into memory

            bmp.Save(screencap_path)               'Save the Screenshot -- this is what might fail
            bmp.Dispose()                   ' Then free up the memory image
            graph.Dispose()                 ' before exiting this load process
            If Logging Then AddLogMessage(4, "CBPrint- capture of screen completed, swapping UI screen back")
        Catch ex As Exception
            MsgBox("error in creation of raw graphic")
            '  if any errors in this routine, report them directly, not in the log
        End Try
        '        Me.Show()  ' this isn't necessary anymore, but it can be handy to check the boundaries during debug
        Swap_back_from_white()      ' restore original colors, whether they've been changed or not -- brute force
        If Logging Then AddLogMessage(4, "CBPrint- entering print dialog")

        ' The printdialog1 class is set up in print_load
        PrintDialog1.PrinterSettings = PrintDocument2.PrinterSettings ' Printdialog1 is the printer selection Dialog
        PrintDialog1.AllowSomePages = False         ' set it to the document settings, then disable a few things
        PrintDialog1.AllowSelection = False
        result = PrintDialog1.ShowDialog()  ' open the print dialog window, and get the results from the user

        If Logging Then AddLogMessage(4, "CBPrint- Return from print dialog")
        If result = DialogResult.OK Then ' only on "print" should we go ahead
            If Logging Then AddLogMessage(4, "CBPrint- Printing accepted, preparing print preview and print")
            PrintDocument2.PrinterSettings = PrintDialog1.PrinterSettings ' set the document settings == printer settings
            PrintPreviewDialog1.ShowDialog() ' open up the print preview, and print from there, if desired
        End If
        Me.Close()                  'close all forms and end the app -- We're DONE!

    End Sub

    Public Sub New()            ' create the form1 structures for later display
        MyBase.New()
        InitializeComponent()

    End Sub
    Private Sub Print_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ' set up values and settings for form1, and establish vectors for the print handler

        PrintFlag = Me.Tag                  ' get the print flag to get user selections
        ' Handle the PrintPage event to write the print logic.
        AddHandler PrintDocument2.PrintPage, AddressOf Me.printDocument2_PrintPage

        ' Specify a PrintDocument instance for the PrintPreviewDialog component.
        Me.PrintPreviewDialog1.Document = Me.PrintDocument2
        ' and a print dialog instance, also
        Me.PrintDialog1.Document = Me.PrintDocument2
        '
        If (PrintFlag And 16) Then  ' If the cancel flag was set in form2, end the application... CANCEL!
            Me.Close()
        End If

    End Sub

    Private Sub printDocument2_PrintPage(ByVal sender As Object, ByVal e As System.Drawing.Printing.PrintPageEventArgs)
        '
        ' For whatever strange reason, an update to .net 3.5 on 2/13/2014 caused this package to begin generating
        ' TWO click interrupts for this routine -- one when the selection area was selected, then yet another when
        ' the print preview print icon was clicked.  It did not do that before the update.  The result was that the
        ' image was scaled on both the first and second passes, resulting in some odd, large, mis-oriented images.
        ' I'm guessing the update fixed a bug which prevented a call to a system routine from generating a
        ' click event that would be passed back to the calling thread... because it sure does now!
        ' 
        ' The code added below concering the variable "firstprint" is an effort to avoid re-scaling the already-scaled
        ' image on the second entry here, and instead to just reload the prior (good) image from the crop file, and
        ' use that for the printer preview.  It works as of 02/14/204...
        '
        '
        '


        Dim temp As Single        ' a temp variable to store sizes while swapping axes for rotation
        Dim scaleratio As Single    ' the best ratio for scaling to the paper size

        Dim pageWidth As Single        'These are the Width and heigth of the current printer's printable area.
        pageWidth = PrintDialog1.PrinterSettings.DefaultPageSettings.PrintableArea.Width

        Dim pageHeight As Single
        pageHeight = PrintDialog1.PrinterSettings.DefaultPageSettings.PrintableArea.Height

        '=================TEST THE PASS COUNTER.  IF THE FIRST PASS, START THE SCALING/RECOLORING PROCESS
        If firstprint = 0 Then      ' on the first pass through, do all the scaling and rotation work, then save it.
            If (PrintFlag And 32) = 0 Then    ' if we're printing a selected area, check for a selection having been made
                If _rRectangle.Width = 0 Or _rRectangle.Height = 0 Then
                    MsgBox("Error!  You must select an area first!")
                    GoTo bad_area_selection
                End If
            End If

            'paint the image to the printpreview control 

            Dim drawImg As New Bitmap(screencap_path)     ' re-loading here; we have the full screen in the drawImg

            If PrintFlag And 32 Then                      ' was "full screen" selected?
                _rRectangle.Width = drawImg.Width         ' if entire screen, just set the values as raw, as-captured
                _rRectangle.Height = drawImg.Height
                _rRectangle.X = 0
                _rRectangle.Y = 0
            End If
            If (PrintFlag And 2) Then                   ' if rotate flag is set
                If (_rRectangle.Width > _rRectangle.Height) Then
                    ' we're going to rotate, so figure out the new origins
                    ' the new Y is just the old x
                    temp = _rRectangle.Y                      ' so preserve the old y origin
                    _rRectangle.Y = _rRectangle.X             ' and set the new Y origin (old x)
                    _rRectangle.X = drawImg.Height - _rRectangle.Height - temp   ' calculate new X origin

                    temp = _rRectangle.Width
                    _rRectangle.Width = _rRectangle.Height
                    _rRectangle.Height = temp                 ' now swap the rectangle height and width before flipping
                    '
                    '
                    drawImg.RotateFlip(RotateFlipType.Rotate90FlipNone)  ' then rotate the actual image 90 degrees

                End If
            End If
            If (PrintFlag And 4) Then                           ' if scale flag is set
                ' Check to see which (width or height) we ought to scale to.
                If (pageWidth / _rRectangle.Width) <= (pageHeight / _rRectangle.Height) Then
                    scaleratio = pageWidth / _rRectangle.Width  ' if the image width exceeds page by more than height
                Else
                    scaleratio = pageHeight / _rRectangle.Height ' if image height exceeds page height by more than width
                End If
            Else
                scaleratio = 1          ' if we're not scaling, just use 1 as a scale multipler
            End If

            ' set up the memory structures for a new set of images
            Dim sourceRectangle As New Rectangle(_rRectangle.X - 5, _rRectangle.Y - 5, _rRectangle.Width + 10, _rRectangle.Height + 10)
            Dim destRectangle1 As New Rectangle(0, 0, _rRectangle.Width * scaleratio, _rRectangle.Height * scaleratio)
            Dim CropImage = New Bitmap(destRectangle1.Width, destRectangle1.Height)

            ' here, crop and scale to a print preview image
            Using grp = Graphics.FromImage(CropImage)
                grp.DrawImage(drawImg, destRectangle1, sourceRectangle, GraphicsUnit.Pixel)
                CropImage.Save(selectcap_path)
            End Using
            '========================================swap to monochrome===============================================
            If PrintFlag And 1 Then     ' if the monochrome flag is set, convert it.
                Dim f3 As Form = New Form3
                Dim X As Integer
                Dim Y As Integer
                Dim clr As Integer
                Dim bm As New Bitmap(CropImage)

                ' OH! This works really well. Don't touch a thing!

                f3.Show()           ' since we're converting to mono, and it takes a while, display the wait msg.
                For X = 0 To bm.Width - 1                   ' for each pixel line
                    For Y = 0 To bm.Height - 1              ' and each pixel in the line
                        ' add the lumas of all the colors (RGB), and average them
                        clr = (CInt(bm.GetPixel(X, Y).R) + _
                               bm.GetPixel(X, Y).G + _
                               bm.GetPixel(X, Y).B) \ 3      ' accumulate luminance figure for this pixel
                        bm.SetPixel(X, Y, Color.FromArgb(clr, clr, clr))  ' create a gray pixel with this luma
                    Next Y  ' next pixel
                Next X ' next line of pixels

                drawImg = bm                    ' drawImg is what we process below, so get the bm image into drawImg
                drawImg.Save(selectcap_path)    ' save the file to disk
                ' and because we don't want to re-scale it AGAIN, use the dimensions of CropImage as the rectangles
                destRectangle1 = New Rectangle(0, 0, CropImage.Width, CropImage.Height)
                sourceRectangle = New Rectangle(0, 0, CropImage.Width, CropImage.Height)
                f3.Close()                      ' done remapping to mono, so take down the wait message
            End If
            '=============================================end mono swap===================================================
            e.Graphics.DrawImage(drawImg, destRectangle1, sourceRectangle, GraphicsUnit.Pixel)
            ' set the 'e' value of this event to the memory copy of the image

            drawImg.Dispose()           ' then get rid of the memory versions
            CropImage.Dispose()
            '====================================END OF FIRST-PASS PROCESSING
        Else               ' On the SECOND entry here, just re-load the already-scaled work, and print it.
            '====================================BEGIN THE SECOND-PASS PROCESS
            Dim CropImage = New Bitmap(selectcap_path)      ' Just get the already-done work
            Dim croprectangle = New Rectangle(0, 0, CropImage.Width, CropImage.Height)
            e.Graphics.DrawImage(CropImage, croprectangle, croprectangle, GraphicsUnit.Pixel)
        End If
        '====================================END OF SECOND-PASS PROCESS

        firstprint += 1  ' on whichever pass, bump the first-pass counter.  Subsequent requests to print
        '                  will see this as non-zero, and process as if the second pass.
bad_area_selection:
    End Sub
    Private Sub Form1_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        ' enter here to select the starting point of the new rubber frame.

        Me._bRubberBandingOn = Not _bRubberBandingOn

        If Me._bRubberBandingOn Then
            If _pClickStart = Nothing Then _pClickStart = New Point

            _pClickStart.X = e.X
            _pClickStart.Y = e.Y
            _pNow.X = e.X
            _pNow.Y = e.Y
        End If
        Me.Invalidate()
    End Sub

    Private Sub Form1_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        ' extend the rubber frame 
        If Me._bRubberBandingOn Then
            If _pNow = Nothing Then _pNow = New Point
            Me._pNow.X = e.X
            Me._pNow.Y = e.Y
            Me.Invalidate()
        End If


    End Sub
    Private Sub Form1_MouseHover(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseHover
        '
        ' this is a cheat to make the "full screen" option use all the same code as the "selected area" option.
        ' This routine detects when form1 is finally in context and has mouse control.
        ' As soon as it does, IF the print flag says to print all, then it simulates an enter key pressed to start
        ' the print capture process, without the user having to do anything.
        '
        If PrintFlag And 32 Then
            SendKeys.Send(Chr(13)) ' fake that the enter key was pressed
        End If
    End Sub
    Private Sub Form1_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        ' terminate growing the rubber rectangle
        Me._bRubberBandingOn = Not Me._bRubberBandingOn
        If Not Me._bRubberBandingOn Then
            If _pClickStop = Nothing Then _pClickStop = New Point
            _pClickStop.X = e.X
            _pClickStop.Y = e.Y
            Me.Invalidate()
        End If
    End Sub
    Private Sub Form1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        ' Inform the user that selection is now in process, and how to do it.  Then it creates the rubber shape.
        ' this routine calls DrawString, which shows how to PAINT characters in a form, instead of using text boxes.

        Dim drawtext As String
        drawtext = TextTranslation.Translate("Left-click-drag to select area." _
        + vbCrLf + "Press Enter when ready to print.")

        Dim _penNew As New Pen(Color.Black, 2)
        _rRectangle.X = _pClickStart.X
        _rRectangle.Y = _pClickStart.Y

        If Me._bRubberBandingOn Then
            _rRectangle.Width = Me._pNow.X - _pClickStart.X
            _rRectangle.Height = Me._pNow.Y - _pClickStart.Y
        Else '-- else if we are done having the shape follow the mouse
            _rRectangle.Width = Me._pClickStop.X - _pClickStart.X
            _rRectangle.Height = Me._pClickStop.Y - _pClickStart.Y
        End If

        _penNew.DashStyle = Drawing2D.DashStyle.DashDot
        e.Graphics.DrawRectangle(_penNew, _rRectangle)

        DrawString(drawtext, Color.Black, 650, 0)         ' post the instructions if we're selecting

    End Sub
    Public Sub DrawString(ByVal gtext As String, ByVal gcolor As Color, ByVal X As Integer, ByVal Y As Integer)
        ' this routine paints text in a form, rather than requiring a text box to display it.

        Dim formGraphics As System.Drawing.Graphics = Me.CreateGraphics()
        Dim drawFont As New System.Drawing.Font(System.Drawing.FontFamily.GenericSerif, 18)
        Dim drawBrush As New  _
           System.Drawing.SolidBrush(gcolor)
        Dim drawFormat As New System.Drawing.StringFormat
        formGraphics.DrawString(gtext, drawFont, drawBrush, _
            X, Y, drawFormat)
        drawFont.Dispose()
        drawBrush.Dispose()
        formGraphics.Dispose()
    End Sub
    Public Sub getCBpath()
        Try                                 ' if this .cb file didn't come from anywhere (is a new, untitled CB file), catch the get path exception
            p = newdoc.Filename             ' that will happen on this line, and create a path to the system temp directory instead
            Dim index1 As Integer = p.LastIndexOf("\"c)     ' extract the path of the current CB file
            rawpath = p.Substring(0, index1 + 1)
        Catch ex As Exception
            rawpath = System.IO.Path.GetTempPath           ' we caught an exception, so find the path to TMP
            If Logging Then
                logstring = "CBprint- CB file has never been saved.  ScreenCapture file will be" + vbCrLf + _
                    "saved to " + rawpath
                AddLogMessage(4, logstring)
            End If
            MsgBox(TextTranslation.Translate("Cambam file has never been saved.  ScreenCapture file will be" + vbCrLf + "saved to ") + rawpath)                      ' and tell the user
        End Try

        screencap_path = rawpath + "CBscreencap.bmp"        ' name of the screen capture file to be saved
        selectcap_path = rawpath + "CBselectcap.bmp"        ' name of the selection capture file to be saved
    End Sub
    Public Sub Swapto_white()

        ' change background to white, and any white layers to the old BG color.  Save everything for restoration later.

        ourlayers = view.CADFile.Layers     ' got to refresh this and the count at run time
        layercount = ourlayers.Count          ' and this
        ReDim colormatrix(layercount)       ' BECAUSE it was already DIM'd as a global.  Now, resize it.
        ReDim Penmatrix(layercount)         ' same as the pen matrix
        panecolor = CamBamConfig.Defaults.ViewBackgroundColor
        If Logging Then
            logstring = System.Drawing.ColorTranslator.ToHtml(panecolor)
            logstring = "CBPrint- Swapto_white subroutine. Drawing Pane BG color is=" + logstring
            AddLogMessage(4, logstring)
        End If


        If layercount > 0 Then
            For i = 0 To layercount - 1
                colormatrix(i) = ourlayers(i).Color
                Penmatrix(i) = ourlayers(i).PenWidth
                ourlayers(i).PenWidth = Our_penwidth ' for printing, use a specific pen width for better dithering

                ' simple test here.  We're going to make the background white.  If the layer color is white,
                ' it gets replaced with the old BG pane color
                If (PrintFlag And 8) Then  ' if "fix colors" selected, then replace 
                    If ((colormatrix(i).ToArgb And &HFFFFFF) = &HFFFFFF) Then   ' if it's white, replace it
                        If Logging Then
                            logstring = ourlayers(i).Name
                            logstring = "CBPrint- layer#" + Str(i + 1) + ", named '" + logstring + "'" + _
                                " was White; swapping to BG color=" + System.Drawing.ColorTranslator.ToHtml(panecolor)
                            AddLogMessage(4, logstring)
                        End If

                        ourlayers(i).Color = panecolor  ' replace the white layer with the old BG color
                    End If
                End If
            Next
        End If

        If Logging Then AddLogMessage(4, "CBPrint- Done with layer color checks, making drawing pane BG white.")
        If (PrintFlag And 8) Then  ' again, if we're fixing colors, change the bg to white
            CamBamConfig.Defaults.ViewBackgroundColor = System.Drawing.Color.White

            view.UpdateViewport() ' this is critical!  RefreshView seems to only update the
            view.RefreshView()    ' drawing area of the screen on SOME machines, but not others!

        End If
    End Sub
    Public Sub Swap_back_from_white()
        If Logging Then AddLogMessage(4, "CBPrint- Swapping colors back to original ones.")
        ReDim Preserve colormatrix(layercount)      ' dimension here for local access, but keep the old contents
        For i = 0 To layercount - 1
            ourlayers(i).Color = colormatrix(i)     ' restore all the colors including those not changed -- brute force fix
            ourlayers(i).PenWidth = Penmatrix(i)     ' restore all the colors including those not changed -- brute force fix
        Next
        CamBamConfig.Defaults.ViewBackgroundColor = panecolor
        view.UpdateViewport()       ' critical!  RefreshView works on other parts of the screen, but on SOME
        view.RefreshView()          ' machines, will not update the drawing area.  Do not eliminate!
    End Sub
End Class

Public Class MyPlugin

    Public Shared Sub InitPlugin(ByRef ui As CamBamUI)
        ' init by adding the print item in the file menu, and setting the handler vector to 'plugin_clicked'
        Dim my1 As New ToolStripMenuItem()
        my1.Text = "Print"
        AddHandler my1.Click, AddressOf plugin_clicked
        ui.Menus.mnuFile.DropDownItems.Insert(6, my1)
        '

    End Sub
    Public Shared Sub plugin_clicked(ByVal sender As Object, ByVal e As EventArgs)

        ' enter here if the "print" item in the file menu was clicked

        Dim ui As CamBamUI = CamBamUI.MainUI
        Dim f2 As Form = New Form2
        Dim fr As Integer

        f2.ShowDialog()             ' tag is initialized to zero; see what it is on return  
        fr = f2.Tag                   ' get the return value with various bits set for options
        Dim f1 As Form = New Form1
        f1.Tag = fr                 ' bit 5= print screen, bit 4= notused,  bit 3= fix colors,
        '                             bit 2= scale, bit 1= rotate, bit 0= notused
        f1.Show()
    End Sub
End Class
